//
//	compiler.cpp
//	Yon
//
//	Created by Miso on 12.8.2009.
//	Copyright 2009 Miso Antonic. All rights reserved.
//
//

#include "compiler.h"
#include "machine.h"


//#include "assert.h"
#include <iostream>

// --- Error

ErrorRef errorNew(wstring s, SourcePosition position)
{
    return new Error(s, position);
}

Error::Error(wstring _description, SourcePosition _position) :
    position(_position),
    description(_description)
{
}

Error::~Error()
{
}

// --- Compiler

#define VERSION_STRING_SIZE 20

// PLATFORM
#define FOLDER_SEPARATOR_CHAR '/'

Compiler::Compiler(MachineRef _machine) :
    compiledVersion(NULL),
    machine(_machine),
    error(NULL)
{
}

CompilerRef compilerNew(const char* path)
{
    return new Compiler(machineNew(path));
}

Compiler::~Compiler()
{
    delete machine;
    delete error;
    free(compiledVersion);
}


// --- Machine Compilation Data Access

#define OBJ_SLOT(obj, slotName) ((obj) == NULL ? NULL : getSlot((obj), symbolForCString(machine->symbolTable, machine->symbolTableLength, slotName)))

ObjectRef getResultObject(MachineRef machine)
{
    ObjectRef module = OBJ_SLOT(machine->objectResult, L"!Module!compiler.yon");
    return OBJ_SLOT(module, L"compilerResult");
}

bool hasCompileError(MachineRef machine)
{
    ObjectRef compileError = OBJ_SLOT(getResultObject(machine), L"error");
    return compileError != NULL && compileError != machine->objectNil;
}

SourcePosition sourcePositionOfCompileError(MachineRef machine)
{
    SourcePosition position = {0, 0, 0, 0, 0};
    ObjectRef compileError = OBJ_SLOT(getResultObject(machine), L"error");
    ObjectRef pos = OBJ_SLOT(compileError, L"pos");
    if (pos == NULL)
        return position;
    
    ObjectRef minLine = OBJ_SLOT(pos, L"minLine");
    ObjectRef minIndex = OBJ_SLOT(pos, L"minIndex");
    ObjectRef maxLine = OBJ_SLOT(pos, L"maxLine");
    ObjectRef maxIndex = OBJ_SLOT(pos, L"maxIndex");
    
    if (minLine == NULL || minIndex == NULL || maxLine == NULL || maxIndex == NULL)
        return position;
    
    position.minLine = (ULONG)REAL(minLine)->value;
    position.minIndex = (ULONG)REAL(minIndex)->value;
    position.maxLine = (ULONG)REAL(maxLine)->value;
    position.maxIndex = (ULONG)REAL(maxIndex)->value;
    return position;
}

CHAR* cStringOfCompileError(MachineRef machine)
{
    ObjectRef compileError = OBJ_SLOT(getResultObject(machine), L"error");
    if (compileError == NULL)
        return NULL;
    StringObjectRef errorMessage = (StringObjectRef)OBJ_SLOT(compileError, L"text");    
    if (errorMessage == NULL)
        return NULL;
    return errorMessage->data;
}

DataObjectRef bytecodeDataOfCompilation(MachineRef machine)
{
    DataObjectRef bytecode = (DataObjectRef)OBJ_SLOT(getResultObject(machine), L"bytecode");
    if ((ObjectRef)bytecode == machine->objectNil)
        bytecode = NULL;
    return bytecode;
}

// --- mark Running

void compilerRun(CompilerRef compiler, const wstring& source, const wstring& sourceName, const wstring& directory, const wstring& library)
{   
    delete compiler->error;
    compiler->error = NULL;
    
    compiler->compiledVersion = (CHAR*)malloc(sizeof(CHAR)*(VERSION_STRING_SIZE+1));
    
    time_t rawtime;
    struct tm * timeinfo;
    time(&rawtime);
    timeinfo = localtime(&rawtime);
    
    wcsftime(compiler->compiledVersion, VERSION_STRING_SIZE+1, L"YC_%y_%m_%d_%H_%M_%S", timeinfo);
    compiler->compiledVersion[VERSION_STRING_SIZE] = NULL;
    
    ObjectRef sourceString = (ObjectRef)newStringObject(compiler->machine, source.c_str(), source.size());
    ObjectRef sourceNameString = (ObjectRef)newStringObject(compiler->machine, sourceName.c_str(), sourceName.size());
    ObjectRef versionString = (ObjectRef)newStringObject(compiler->machine, compiler->compiledVersion, VERSION_STRING_SIZE);
    ObjectRef directoryString = (ObjectRef)newStringObject(compiler->machine, directory.c_str(), directory.size());
    ObjectRef libraryPathString = (ObjectRef)newStringObject(compiler->machine, library.c_str(), library.size());
    
    ObjectRef data[5] = {sourceString, sourceNameString, versionString, directoryString, libraryPathString};
    ObjectRef parameter = (ObjectRef)newArrayObject(compiler->machine, data, 5);
    
    
    machineRun(compiler->machine, parameter);
    
    if (compiler->machine->error)
    {
        ErrorRef error = compiler->machine->error;
        compiler->error = errorNew(error->description, error->position);
    }
    
    
    if (hasCompileError(compiler->machine))
    {
        SourcePosition position = sourcePositionOfCompileError(compiler->machine);
        CHAR* description = cStringOfCompileError(compiler->machine);
        
        compiler->error = errorNew(description, position);
    }
}

// --- mark Getting & Saving results

bool compilerGetResult(CompilerRef compiler, BYTE** data, ULONG* dataSize)
{
    if (compiler->machine->error)
        return false;
    if (hasCompileError(compiler->machine))
        return false;
    
    DataObjectRef dataObject = bytecodeDataOfCompilation(compiler->machine);
    
    *data = dataObject->bytes;
    *dataSize = dataObject->size;
    
    return true;
}

bool compilerSaveResultToFile(CompilerRef compiler, const char* path)
{
    BYTE* data = NULL;
    ULONG dataSize = 0;
    
    if (!compilerGetResult(compiler, &data, &dataSize))
        return false;
    
    FILE* file = fopen(path, "wb");
    if (file == NULL)
        return false;
    fwrite(data, 1, dataSize, file);
    fclose(file);
    
    return true;
}

bool compilerSaveResultWithVersionedFileNameToPath(CompilerRef compiler, const char* path)
{
    ULONG pathSize = strlen(path);
    if (pathSize == 0)
        return false;

    char lastCharOfPath = path[pathSize-1];
    CHAR* version = compiler->compiledVersion;
    ULONG versionSize = VERSION_STRING_SIZE;
    const char* extension = ".ycode";
    ULONG extensionSize = strlen(extension);
    ULONG filePathSize = pathSize + 1 + VERSION_STRING_SIZE + extensionSize + 1;
    char* filePath = (char*)malloc(sizeof(char) * filePathSize);
    char* index = filePath;
    
    
    while (pathSize-- > 0)
        *(index++) = *(path++);

    if (lastCharOfPath != FOLDER_SEPARATOR_CHAR)
        *(index++) = FOLDER_SEPARATOR_CHAR;
    
    while (versionSize-- > 0)
        *(index++) = (char)*(version++);
    
    while (extensionSize-- > 0)
        *(index++) = *(extension++);
    
    *(index++) = NULL;
    
    assert(filePathSize == (index - filePath) || filePathSize == (index - filePath) + 1);
    
    bool result = compilerSaveResultToFile(compiler, filePath);
    free(filePath);
    
    return result;
}
