//
//  module.cpp
//  Yon
//
//  Created by Miso on 12.7.2010.
//  Copyright 2010 Miso Antonic. All rights reserved.
//
//

#include "module.h"
#include "machine.h"
#include "math.h"
#include <iostream>
#include "stdlib.h"
#include "common.h"

#include "input.h"

#define FUNC(name) ObjectRef name(ObjectRef SELF, ObjectRef* parameters, MachineRef machine)
#define PARAMETER(index) (parameters[index-1])

#define ASSERT(invariant, message) \
if (!(invariant))\
{ \
    wstring func(__func__, __func__ + strlen(__func__)); \
    wstring s(func + L" error: " + (message)); \
    machine->internalErrorMessage = s; \
    return machine->objectInternalError; \
}

#define ERROR(message) \
{ \
    wstring s(STRING(message)->data); \
    machine->internalErrorMessage = s; \
    return machine->objectInternalError; \
}

#define CHECK_TYPE(object, expected) ASSERT((object)->type == (expected), L"" #object " is not " #expected)

#define MAKE_STRING(buffer, length) ((ObjectRef)newStringObject(&machine->objects, machine->pool, buffer, length))
#define MAKE_STRING_DONT_COPY(buffer, length) ((ObjectRef)newStringObjectDontCopy(&machine->objects, machine->pool, buffer, length))
#define MAKE_REAL(value) ((ObjectRef)newRealObject(&machine->objects, machine->pool, value))
#define MAKE_ARRAY(buffer, length) ((ObjectRef)newArrayObject(&machine->objects, machine->pool, buffer, length))
#define MAKE_DATA(bytes, length) ((ObjectRef)newDataObject(&machine->objects, machine->pool, bytes, length))

FUNC(real_string)
{
	CHECK_TYPE(PARAMETER(1), typeReal);
	DOUBLE x = REAL(PARAMETER(1))->value;

	DOUBLE zero = 0.0;
	DOUBLE one = 1.0;

	DOUBLE NaN = zero/zero;
	DOUBLE INF = one/zero;
	DOUBLE NEG_INF = -one/zero;
	
	if (isnan(x))
		return MAKE_STRING((CHAR*)L"NAN", 3);
	if (x == INF)
		return MAKE_STRING((CHAR*)L"OO", 3);
	if (x == -INF)
		return MAKE_STRING((CHAR*)L"-OO", 3);

	CHAR buffer[32];
	BYTE length = swprintf(buffer, 32, L"%g", x);
	return MAKE_STRING(buffer, length);
}

FUNC(function_string)
{	
	CHECK_TYPE(PARAMETER(1), typeFunction);
	CHAR buffer[255];
	BYTE length = swprintf(buffer, 255, L"Function Object, %d parameters, bytecode address: %04d", FUNCTION(PARAMETER(1))->numberOfParameters,  FUNCTION(PARAMETER(1))->startAddress - machine->bytecodes);
	return MAKE_STRING(buffer, length);
}

FUNC(nativeFunction_string)
{
	CHECK_TYPE(PARAMETER(1), typeNativeFunction);
	CHAR buffer[255];
	BYTE length = swprintf(buffer, 255, L"Native Function Object, %d parameters, address: 0x%x", NATIVE_FUNCTION(PARAMETER(1))->numberOfParameters,  NATIVE_FUNCTION(PARAMETER(1))->address);
	return MAKE_STRING(buffer, length);
}

FUNC(printToConsole)
{
	CHECK_TYPE(PARAMETER(1), typeString);
	if (STRING(PARAMETER(1))->data)
		wprintf(L"%ls\n", STRING(PARAMETER(1))->data);
	else
		wprintf(L"''\n");
    fflush(stdout);
	return machine->objectNil;
}

FUNC(array_size)
{
	CHECK_TYPE(PARAMETER(1), typeArray);
	return MAKE_REAL(ARRAY(PARAMETER(1))->size);
}

FUNC(array_item)
{
	CHECK_TYPE(SELF, typeArray);
	CHECK_TYPE(PARAMETER(1), typeReal);
	ASSERT(0 <= REAL(PARAMETER(1))->value && REAL(PARAMETER(1))->value < ARRAY(SELF)->size, L"Runtime error, Array index out of range.");
	return (ObjectRef)ARRAY(SELF)->data[(ULONG)REAL(PARAMETER(1))->value];
}

FUNC(array_setItem)
{
	CHECK_TYPE(SELF, typeArray);
	CHECK_TYPE(PARAMETER(1), typeReal);
	ASSERT(0 <= REAL(PARAMETER(1))->value && REAL(PARAMETER(1))->value < ARRAY(SELF)->size, L"Runtime error, Array index out of range.");
    ARRAY(SELF)->data[(ULONG)REAL(PARAMETER(1))->value] = PARAMETER(2);
	return PARAMETER(2);
}

FUNC(array_plus)
{
	CHECK_TYPE(SELF, typeArray);
	CHECK_TYPE(PARAMETER(1), typeArray);
    ULONG length = ARRAY(SELF)->size + ARRAY(PARAMETER(1))->size;


    ObjectRef result = MAKE_ARRAY(NULL, length);
    
    
    memcpy(&ARRAY(result)->data[0], ARRAY(SELF)->data, sizeof(ObjectRef)*ARRAY(SELF)->size);
    memcpy(&ARRAY(result)->data[ARRAY(SELF)->size], ARRAY(PARAMETER(1))->data, sizeof(ObjectRef)*ARRAY(PARAMETER(1))->size);
    
    return result;
}

FUNC(array_add)
{
	CHECK_TYPE(SELF, typeArray);
    ARRAY(SELF)->size += 1;
    REALLOC_ARRAY(ARRAY(SELF)->data, ObjectRef, ARRAY(SELF)->size);
    ARRAY(SELF)->data[ARRAY(SELF)->size-1] = PARAMETER(1);
    return SELF;
}

FUNC(array_append)
{
	CHECK_TYPE(SELF, typeArray);
	CHECK_TYPE(PARAMETER(1), typeArray);
    ULONG oldLength = ARRAY(SELF)->size;
    REALLOC_ARRAY(ARRAY(SELF)->data, ObjectRef, oldLength + ARRAY(PARAMETER(1))->size);
    memcpy(&ARRAY(SELF)->data[oldLength], ARRAY(PARAMETER(1))->data, sizeof(ObjectRef)*ARRAY(PARAMETER(1))->size);
    ARRAY(SELF)->size += ARRAY(PARAMETER(1))->size;
    return SELF;
}

FUNC(array_subarray)
{
	CHECK_TYPE(SELF, typeArray);
	CHECK_TYPE(PARAMETER(1), typeReal);
	CHECK_TYPE(PARAMETER(2), typeReal);
    ULONG from = (ULONG)MAX(0, round(REAL(PARAMETER(1))->value));
    ULONG to = (ULONG)MIN(STRING(SELF)->size-1, REAL(PARAMETER(2))->value);
    ASSERT(from <= to, L"Runtime error, first parameter of array slicing must be less or equal than second");
    
    ObjectRef result = MAKE_ARRAY(NULL, to-from+1);
    memcpy(&ARRAY(result)->data[0], &(ARRAY(SELF)->data[from]), sizeof(ObjectRef)*(ARRAY(result)->size));
    
    return result;
}

FUNC(string_number)
{
	CHECK_TYPE(PARAMETER(1), typeString);
    CHAR* restrict = NULL;
    StringObjectRef s = STRING(PARAMETER(1));
	DOUBLE x = wcstod(s->data, &restrict);
    if (restrict == s->data)
		return machine->objectNil;
    return MAKE_REAL(x);
}

FUNC(string_contains)
{
	CHECK_TYPE(SELF, typeString);
	CHECK_TYPE(PARAMETER(1), typeString);

	CHAR* search = wcsstr(STRING(SELF)->data, STRING(PARAMETER(1))->data);

	if (search == NULL)
		return machine->objectNo;
	else
		return machine->objectYes;
}

FUNC(string_writeToFile)
{
	CHECK_TYPE(SELF, typeString);
	CHECK_TYPE(PARAMETER(1), typeString);
    
    StringObjectRef s = STRING(SELF);
    char path[2048];
    
    
    ASSERT(wcstombs(path, STRING(PARAMETER(1))->data, 2048) < 2048, L"Runtime error, Could not convert path to narrow chars");
    
    FILE* file = fopen(path, "wb");
    ASSERT(file != NULL, L"Runtime error, Could not open file.");

	char* cs = NULL;
	NEW_ARRAY(cs, char, s->size+1);
	cs[s->size] = NULL;

	for (ULONG i=0; i<s->size; i++)
		cs[i] = s->data[i];

	fwrite(cs, 1, s->size, file);
    
	fclose(file);
    return SELF;
}

FUNC(string_item)
{
    CHECK_TYPE(SELF, typeString);
    CHECK_TYPE(PARAMETER(1), typeReal);
	ASSERT(0 <= REAL(PARAMETER(1))->value && REAL(PARAMETER(1))->value < STRING(SELF)->size, L"Runtime error, String index out of range.");
    return MAKE_STRING(&(STRING(SELF)->data[(ULONG)REAL(PARAMETER(1))->value]), 1);
}

FUNC(string_setItem)
{
	CHECK_TYPE(SELF, typeString);
	CHECK_TYPE(PARAMETER(1), typeReal);
	CHECK_TYPE(PARAMETER(2), typeString);

    ASSERT(STRING(PARAMETER(2))->size == 1, L"Runtime error, The string that is being assigned into other string must be a single character.");
    
	ASSERT(0 <= REAL(PARAMETER(1))->value && REAL(PARAMETER(1))->value < STRING(SELF)->size, L"Runtime error, String index out of range.");
    
    STRING(SELF)->data[(ULONG)REAL(PARAMETER(1))->value] = STRING(PARAMETER(2))->data[0];
    return PARAMETER(2);
}

FUNC(string_size)
{
	CHECK_TYPE(PARAMETER(1), typeString);
    return MAKE_REAL(STRING(PARAMETER(1))->size);
}

FUNC(string_concat)
{
	CHECK_TYPE(SELF, typeString);
	CHECK_TYPE(PARAMETER(1), typeString);

    ULONG length = STRING(SELF)->size + STRING(PARAMETER(1))->size;
    ObjectRef result = MAKE_STRING(NULL, length);
    
    wcsncpy(STRING(result)->data, STRING(SELF)->data, STRING(SELF)->size);
    wcsncpy(&STRING(result)->data[STRING(SELF)->size], STRING(PARAMETER(1))->data, STRING(PARAMETER(1))->size);
    return result;
}

#define MAX(a, b) ((a) > (b) ? (a) : (b))
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define ABS(a)    (((a) >= 0) ? (a) : (-(a)))

FUNC(string_substring)
{
	CHECK_TYPE(SELF, typeString);
	CHECK_TYPE(PARAMETER(1), typeReal);
	CHECK_TYPE(PARAMETER(2), typeReal);
    LONG from = (LONG)MAX(0, round(REAL(PARAMETER(1))->value));
    LONG to = (LONG)MIN(STRING(SELF)->size-1, REAL(PARAMETER(2))->value);
    ASSERT(from <= to, L"Runtime error, from > to in substring.");
    
    ObjectRef result = MAKE_STRING(NULL, to-from+1);
    wcsncpy(STRING(result)->data, &(STRING(SELF)->data[from]), to-from+1);
    
    return result;
}

FUNC(string_indexOfSubstring)
{
	CHECK_TYPE(SELF, typeString);
	CHECK_TYPE(PARAMETER(1), typeString);
	ASSERT(PARAMETER(2) == machine->objectNil || PARAMETER(2)->type == typeReal, L"Runtime error, Second parameter of indexOfSubstring must be Number or nil.");
	if (PARAMETER(2)->type == typeReal)
		ASSERT(REAL(PARAMETER(2))->value <= STRING(SELF)->size, L"Runtime error, Second parameter of indexOfSubstring cannot be larger than the size of the string.");

	ULONG offset = PARAMETER(2) == machine->objectNil ? 0 : REAL(PARAMETER(2))->value;

	CHAR* search = wcsstr(STRING(SELF)->data+offset, STRING(PARAMETER(1))->data);

	if (search == NULL)
		return machine->objectNil;

	return MAKE_REAL(search - STRING(SELF)->data);
}

FUNC(string_equals)
{
	CHECK_TYPE(SELF, typeString);
    if (SELF->type != PARAMETER(1)->type)
        return machine->objectNo;
    else
        return STRING(SELF)->size == STRING(PARAMETER(1))->size && wcscmp(STRING(SELF)->data, STRING(PARAMETER(1))->data) == 0 ? machine->objectYes : machine->objectNo;
}

FUNC(string_isLess)
{
	CHECK_TYPE(SELF, typeString);
	CHECK_TYPE(PARAMETER(1), typeString);
    return wcscmp(STRING(SELF)->data, STRING(PARAMETER(1))->data) < 0 ? machine->objectYes : machine->objectNo;
}

FUNC(string_append)
{
	CHECK_TYPE(SELF, typeString);
	CHECK_TYPE(PARAMETER(1), typeString);
    ULONG oldLength = STRING(SELF)->size;
    ULONG newLength = oldLength + STRING(PARAMETER(1))->size;
    REALLOC_ARRAY(STRING(SELF)->data, CHAR, newLength + 1);
    memcpy(&STRING(SELF)->data[oldLength], STRING(PARAMETER(1))->data, sizeof(CHAR)*STRING(PARAMETER(1))->size);
    STRING(SELF)->size = newLength;
    STRING(SELF)->data[newLength] = NULL;
    return SELF;
}

FUNC(function_isFrozen)
{
	CHECK_TYPE(PARAMETER(1), typeFunction);
	return (ObjectRef)FUNCTION(PARAMETER(1))->frozenCall ? machine->objectYes : machine->objectNo;
}


FUNC(bool_and)
{
	CHECK_TYPE(SELF, typeBool);
	CHECK_TYPE(PARAMETER(1), typeBool);
	if (BOOL(SELF)->value && BOOL(PARAMETER(1))->value)
		return machine->objectYes;
	else
		return machine->objectNo;
}

FUNC(bool_or)
{
	CHECK_TYPE(SELF, typeBool);
	CHECK_TYPE(PARAMETER(1), typeBool);
	if (BOOL(SELF)->value || BOOL(PARAMETER(1))->value)
		return machine->objectYes;
	else
		return machine->objectNo;
}

FUNC(bool_xor)
{
	CHECK_TYPE(SELF, typeBool);
	CHECK_TYPE(PARAMETER(1), typeBool);
	if (BOOL(SELF)->value ^ BOOL(PARAMETER(1))->value)
		return machine->objectYes;
	else
		return machine->objectNo;
}

FUNC(bool_not)
{
	CHECK_TYPE(SELF, typeBool);
	if (BOOL(SELF)->value)
		return machine->objectNo;
	else
		return machine->objectYes;
}

FUNC(real_arrayTo)
{
	CHECK_TYPE(SELF, typeReal);
	CHECK_TYPE(PARAMETER(1), typeReal);
	LONG a = (LONG)round(REAL(SELF)->value);
	LONG b = (LONG)round(REAL(PARAMETER(1))->value);
	LONG step = a < b ? 1 : -1;
	ULONG length = ABS(b - a) + 1;
	DOUBLE value = b;
	
	ObjectRef result = MAKE_ARRAY(NULL, length);
	
	while (length != 0)
	{
		--length;
		ARRAY(result)->data[length] = MAKE_REAL(value);
		value -= step;
	}
	return result;
}

FUNC(real_equals)
{
	CHECK_TYPE(SELF, typeReal);
    if (SELF->type != PARAMETER(1)->type)
        return machine->objectNo;
    else
        return REAL(SELF)->value == REAL(PARAMETER(1))->value ? machine->objectYes : machine->objectNo;
}

FUNC(real_isLess)
{
	CHECK_TYPE(SELF, typeReal);
	CHECK_TYPE(PARAMETER(1), typeReal);
    return REAL(SELF)->value < REAL(PARAMETER(1))->value ? machine->objectYes : machine->objectNo;
}

FUNC(real_unaryminus)
{
	CHECK_TYPE(SELF, typeReal);
    return MAKE_REAL(-REAL(SELF)->value);
}

FUNC(real_plus)
{
	CHECK_TYPE(SELF, typeReal);
	CHECK_TYPE(PARAMETER(1), typeReal);
    return MAKE_REAL(REAL(SELF)->value + REAL(PARAMETER(1))->value);
}

FUNC(real_minus)
{
	CHECK_TYPE(SELF, typeReal);
	CHECK_TYPE(PARAMETER(1), typeReal);
    return MAKE_REAL(REAL(SELF)->value - REAL(PARAMETER(1))->value);
}

FUNC(real_multiply)
{
	CHECK_TYPE(SELF, typeReal);
	CHECK_TYPE(PARAMETER(1), typeReal);
    return MAKE_REAL(REAL(SELF)->value * REAL(PARAMETER(1))->value);
}

FUNC(real_divide)
{
	CHECK_TYPE(SELF, typeReal);
	CHECK_TYPE(PARAMETER(1), typeReal);
    return MAKE_REAL(REAL(SELF)->value / REAL(PARAMETER(1))->value);
}

FUNC(real_mod)
{
	CHECK_TYPE(SELF, typeReal);
	CHECK_TYPE(PARAMETER(1), typeReal);
    return MAKE_REAL((ULONG)(REAL(SELF)->value) % (ULONG)(REAL(PARAMETER(1))->value));
}

FUNC(real_div)
{
	CHECK_TYPE(SELF, typeReal);
	CHECK_TYPE(PARAMETER(1), typeReal);
    return MAKE_REAL((ULONG)(REAL(SELF)->value) / (ULONG)(REAL(PARAMETER(1))->value));
}

FUNC(real_round)
{
	CHECK_TYPE(PARAMETER(1), typeReal);
    return MAKE_REAL(round(REAL(PARAMETER(1))->value));
}

FUNC(real_floor)
{
	CHECK_TYPE(PARAMETER(1), typeReal);
    return MAKE_REAL(floor(REAL(PARAMETER(1))->value));
}

FUNC(real_ceil)
{
	CHECK_TYPE(PARAMETER(1), typeReal);
    return MAKE_REAL(ceil(REAL(PARAMETER(1))->value));
}

FUNC(real_abs)
{
	CHECK_TYPE(PARAMETER(1), typeReal);
    return MAKE_REAL(ABS(REAL(PARAMETER(1))->value));
}

FUNC(real_sin)
{
	CHECK_TYPE(PARAMETER(1), typeReal);
    return MAKE_REAL(sin(REAL(PARAMETER(1))->value));
}

FUNC(real_cos)
{
	CHECK_TYPE(PARAMETER(1), typeReal);
    return MAKE_REAL(cos(REAL(PARAMETER(1))->value));
}

FUNC(real_log)
{
	CHECK_TYPE(PARAMETER(1), typeReal);
	CHECK_TYPE(PARAMETER(2), typeReal);

	DOUBLE a = log(REAL(PARAMETER(1))->value);
	DOUBLE b = log(REAL(PARAMETER(2))->value);

    return MAKE_REAL(a/b);
}

FUNC(real_exp)
{
	CHECK_TYPE(SELF, typeReal);
	CHECK_TYPE(PARAMETER(1), typeReal);
    return MAKE_REAL(pow(REAL(SELF)->value, REAL(PARAMETER(1))->value));
}

FUNC(real_sqrt)
{
	CHECK_TYPE(PARAMETER(1), typeReal);
    return MAKE_REAL(sqrt(REAL(PARAMETER(1))->value));
}

FUNC(createempty)
{
	return newObjectFromPool(&machine->objects, machine->pool, typeObject, sizeof(Object));
}

FUNC(setPrimitivePrototype)
{
	CHECK_TYPE(PARAMETER(1), typeReal);
	int type = REAL(PARAMETER(1))->value;
	ASSERT(type >= 0 && type < numberOfObjectTypes, L"No such primitive type");
	machine->primitivePrototypes[type] = PARAMETER(2);
	return machine->objectNil;
}

FUNC(object_slots)
{	
    ULONG slotCount = PARAMETER(1)->slotCount;
	ObjectRef result = MAKE_ARRAY(NULL, slotCount);
	
	while (slotCount != 0)
	{
		--slotCount;
		ARRAY(result)->data[slotCount] = MAKE_STRING(machine->symbolTable[PARAMETER(1)->slots[slotCount].symbol], wcslen(machine->symbolTable[PARAMETER(1)->slots[slotCount].symbol]));
	}
	return result;
}

FUNC(object_slotValues)
{	
    ULONG slotCount = PARAMETER(1)->slotCount;
	ObjectRef result = MAKE_ARRAY(NULL, slotCount);
	
	while (slotCount != 0)
	{
		--slotCount;
		ARRAY(result)->data[slotCount] = PARAMETER(1)->slots[slotCount].value;
	}
	return result;
}

FUNC(object_equals)
{
    if (SELF == PARAMETER(1))
        return machine->objectYes;
    else
        return machine->objectNo;
}

FUNC(stringFromFile)
{
	CHECK_TYPE(PARAMETER(1), typeString);
    
    char path[1024];
    
    
    ASSERT(wcstombs(path, STRING(PARAMETER(1))->data, 1024) < 1024, L"Runtime error, Could not convert path to narrow chars");
    
    FILE* file = fopen(path, "r");
    ASSERT(file != NULL, L"Runtime error, Could not open file.");
    
	CHAR* source = NULL;
	ULONG numberRead = 0, index = 0;
	
	do
	{
		REALLOC_ARRAY(source, CHAR, index + 1);
		char c;
		numberRead = fread(&c, 1, 1, file);
		source[index] = c;
		
		index += numberRead;
	} while (numberRead);
	REALLOC_ARRAY(source, CHAR, index + 1);
	source[index] = 0; // NULL terminate
	
	fclose(file);
    return MAKE_STRING_DONT_COPY(source, index);
}

FUNC(stringFromInput)
{
	CHECK_TYPE(PARAMETER(1), typeString);

	CHAR str[1024];

	showInput(STRING(PARAMETER(1))->data, str, 1024);


	/*
    CHAR str[1024];

	if (fgetws(str, 1024, stdin) == NULL)
	{
		str[0] = NULL;
	}
    
	ULONG i = 0;
	while (str[i] != NULL)
	{
		if (str[i] == '\n')
			str[i] = NULL;
		i++;
	}*/

	return MAKE_STRING(str, wcslen(str));
}

FUNC(fileExists)
{
	CHECK_TYPE(PARAMETER(1), typeString);
    
    char path[2048];
    
    
    ASSERT(wcstombs(path, STRING(PARAMETER(1))->data, 2048) < 2048, L"Runtime error, Could not convert path to narrow chars");
    
    FILE* file = fopen(path, "r");
    
    if (file == NULL)
        return machine->objectNo;
    
    fclose(file);
    return machine->objectYes;
}

FUNC(stringFromLibrary)
{    
    FILE* file = fopen("/Users/miso/Sites/lib.yon", "r");
    ASSERT(file != NULL, L"Runtime error, Could not open file.");
    
	CHAR* source = NULL;
	ULONG numberRead = 0, index = 0;
	
	do
	{
		REALLOC_ARRAY(source, CHAR, index + 1);
		char c;
		numberRead = fread(&c, 1, 1, file);
		source[index] = c;
		
		index += numberRead;
	} while (numberRead);
	REALLOC_ARRAY(source, CHAR, index + 1);
	source[index] = 0; // NULL terminate
	
	fclose(file);
    return MAKE_STRING_DONT_COPY(source, index);
}

FUNC(charFromNumber)
{
	CHECK_TYPE(PARAMETER(1), typeReal);
    CHAR c[2] = {(LONG)(REAL(PARAMETER(1))->value), 0};
    return MAKE_STRING(c, 1);
}

FUNC(numberFromChar)
{
	CHECK_TYPE(PARAMETER(1), typeString);
	ASSERT(STRING(PARAMETER(1))->size == 1, L"Runtime error, The parameter string must be a single character.");
	return MAKE_REAL((DOUBLE)STRING(PARAMETER(1))->data[0]);
}

FUNC(charEof)
{
    CHAR eof[2] = {EOF, 0};
    return MAKE_STRING(eof, 1);
}

FUNC(charNewLine)
{
    return MAKE_STRING((CHAR*)L"\n", 1);
}

FUNC(charTab)
{
    return MAKE_STRING((CHAR*)L"\t", 1);
}

FUNC(boolYes)
{
    return machine->objectYes;
}

FUNC(boolNo)
{
    return machine->objectNo;
}

FUNC(error)
{
	CHECK_TYPE(PARAMETER(1), typeString);
    ERROR(PARAMETER(1));
}

FUNC(getParameterObject)
{
    return machine->objectParameter;
}

FUNC(data_size)
{
	CHECK_TYPE(PARAMETER(1), _typeData);
    
    return MAKE_REAL((DOUBLE)DATA(PARAMETER(1))->size);
}

#define DATA_APPEND_VALUE(t, v) \
    DataObjectRef data = DATA(SELF); \
    t value = (t)(v); \
    ULONG size = sizeof(t); \
	REALLOC_ARRAY(data->bytes, BYTE, data->size+size); \
    memcpy(&data->bytes[data->size], &value, size); \
    data->size += size;

FUNC(data_appendByte)
{
	CHECK_TYPE(SELF, _typeData);
	CHECK_TYPE(PARAMETER(1), typeReal);
    
    DATA_APPEND_VALUE(BYTE, REAL(PARAMETER(1))->value);
    
    return SELF;
}

FUNC(data_appendUlong)
{
	CHECK_TYPE(SELF, _typeData);
	CHECK_TYPE(PARAMETER(1), typeReal);
    
    DATA_APPEND_VALUE(ULONG, REAL(PARAMETER(1))->value);
    
    return SELF;
}

FUNC(data_appendDouble)
{
	CHECK_TYPE(SELF, _typeData);
	CHECK_TYPE(PARAMETER(1), typeReal);
    
    DATA_APPEND_VALUE(DOUBLE, REAL(PARAMETER(1))->value);
    
    return SELF;
}

FUNC(data_appendBool)
{
	CHECK_TYPE(SELF, _typeData);
	CHECK_TYPE(PARAMETER(1), typeBool);
    
    DATA_APPEND_VALUE(bool, PARAMETER(1) == machine->objectYes);
    
    return SELF;
}

FUNC(data_appendString)
{
	CHECK_TYPE(SELF, _typeData);
	CHECK_TYPE(PARAMETER(1), typeString);
    
    DataObjectRef data = DATA(SELF);
    StringObjectRef s = STRING(PARAMETER(1));
    ULONG size = s->size * sizeof(CHAR32);
    
	
	REALLOC_ARRAY(data->bytes, BYTE, data->size+size);

	for (ULONG i=0; i<s->size; i++)
		*(((CHAR32*)(&data->bytes[data->size])) + i) = s->data[i];

    data->size += size;
    
    return SELF;
}

FUNC(data_appendData)
{
	CHECK_TYPE(SELF, _typeData);
	CHECK_TYPE(PARAMETER(1), _typeData);
    
    DataObjectRef data = DATA(SELF);
    DataObjectRef data2 = DATA(PARAMETER(1));
    
	REALLOC_ARRAY(data->bytes, BYTE, data->size + data2->size);

    memcpy(&data->bytes[data->size], &data2->bytes[0], data2->size);
    data->size += data2->size;
    
    return SELF;
}

FUNC(data_rewriteUlong)
{
	CHECK_TYPE(SELF, _typeData);
	CHECK_TYPE(PARAMETER(1), typeReal);
	CHECK_TYPE(PARAMETER(2), typeReal);
    
    DataObjectRef data = DATA(SELF);
    ULONG position = (ULONG)round(REAL(PARAMETER(1))->value);
    ULONG value = (ULONG)round(REAL(PARAMETER(2))->value);
    
    ASSERT(position + sizeof(ULONG) <= data->size, L"Runtime error, The position specified for the data_rewriteUlong method is out of range");
    
    memcpy(&data->bytes[position], &value, sizeof(ULONG));
    return SELF;
}

FUNC(data_item)
{
	CHECK_TYPE(SELF, _typeData);
	CHECK_TYPE(PARAMETER(1), typeReal);
    
    DataObjectRef data = DATA(SELF);
    ULONG position = (ULONG)round(REAL(PARAMETER(1))->value);
    ASSERT(position <= data->size, L"Runtime error, The position specified for the data_item method is out of range");
    
    BYTE value = data->bytes[position];
    
    return MAKE_REAL(value);
}

FUNC(data_writeToFile)
{
	CHECK_TYPE(SELF, _typeData);
	CHECK_TYPE(PARAMETER(1), typeString);
    
    
    DataObjectRef data = DATA(SELF);
    char path[2048];
    
    
    ASSERT(wcstombs(path, STRING(PARAMETER(1))->data, 2048) < 2048, L"Runtime error, Could not convert path to narrow chars");
    
    FILE* file = fopen(path, "wb");
    ASSERT(file != NULL, L"Runtime error, Could not open file.");
    
    fwrite (data->bytes, 1, data->size, file);
    
	fclose(file);
    return SELF;

}

FUNC(data_new)
{
    return MAKE_DATA(NULL, 0);
}


#define EXPORT_FUNCTION(name, count, wantsself) if (wcscmp(L###name, functionName) == 0) { *function = (NativeFunctionRef)name; *parameterCount = count; *needsCurriedSelf = wantsself; return true; }
bool getNativeFunction(CHAR* functionName, NativeFunctionRef* function, BYTE* parameterCount, bool* needsCurriedSelf)
{	
	EXPORT_FUNCTION(createempty, 0, false);

	EXPORT_FUNCTION(real_string, 1, false);
	EXPORT_FUNCTION(function_string, 1, false);
	EXPORT_FUNCTION(nativeFunction_string, 1, false);
	EXPORT_FUNCTION(printToConsole, 1, false);
	EXPORT_FUNCTION(array_size, 1, false);
	EXPORT_FUNCTION(array_item, 1, true);
	EXPORT_FUNCTION(array_plus, 1, true);
	EXPORT_FUNCTION(array_setItem, 2, true);
    EXPORT_FUNCTION(array_append, 1, true);
    EXPORT_FUNCTION(array_add, 1, true);
    EXPORT_FUNCTION(array_subarray, 2, true);
	EXPORT_FUNCTION(function_isFrozen, 1, false);

	EXPORT_FUNCTION(bool_and, 1, true)
	EXPORT_FUNCTION(bool_or, 1, true)
	EXPORT_FUNCTION(bool_xor, 1, true)
	EXPORT_FUNCTION(bool_not, 0, true)

	EXPORT_FUNCTION(real_arrayTo, 1, true);
	EXPORT_FUNCTION(real_equals, 1, true);
	EXPORT_FUNCTION(real_isLess, 1, true);
	EXPORT_FUNCTION(real_exp, 1, true);
	EXPORT_FUNCTION(real_sqrt, 1, false);

	EXPORT_FUNCTION(real_unaryminus, 0, true);

	EXPORT_FUNCTION(real_plus, 1, true)
	EXPORT_FUNCTION(real_minus, 1, true)
	EXPORT_FUNCTION(real_multiply, 1, true)
	EXPORT_FUNCTION(real_divide, 1, true)

	EXPORT_FUNCTION(real_mod, 1, true)
	EXPORT_FUNCTION(real_div, 1, true)
	EXPORT_FUNCTION(real_round, 1, false)
	EXPORT_FUNCTION(real_floor, 1, false)
	EXPORT_FUNCTION(real_ceil, 1, false)
	EXPORT_FUNCTION(real_abs, 1, false)
	EXPORT_FUNCTION(real_sin, 1, false)
	EXPORT_FUNCTION(real_cos, 1, false)
	EXPORT_FUNCTION(real_log, 2, false)

    
    EXPORT_FUNCTION(string_item, 1, true);
    EXPORT_FUNCTION(string_size, 1, false);
	EXPORT_FUNCTION(string_concat, 1, true);
	EXPORT_FUNCTION(string_substring, 2, true);
	EXPORT_FUNCTION(string_indexOfSubstring, 2, true);
	EXPORT_FUNCTION(string_setItem, 2, true);
	EXPORT_FUNCTION(string_equals, 1, true);
	EXPORT_FUNCTION(string_isLess, 1, true);
	EXPORT_FUNCTION(string_contains, 1, true);
    EXPORT_FUNCTION(string_append, 1, true);
    EXPORT_FUNCTION(string_number, 1, false);
	EXPORT_FUNCTION(string_writeToFile, 1, true);
    
    
    EXPORT_FUNCTION(data_size, 1, false);
    EXPORT_FUNCTION(data_appendByte, 1, true)
    EXPORT_FUNCTION(data_appendUlong, 1, true)
    EXPORT_FUNCTION(data_appendDouble, 1, true)
    EXPORT_FUNCTION(data_appendBool, 1, true)
    EXPORT_FUNCTION(data_appendString, 1, true)
    EXPORT_FUNCTION(data_appendData, 1, true)    
    EXPORT_FUNCTION(data_rewriteUlong, 2, true)
    EXPORT_FUNCTION(data_writeToFile, 1, true)
    EXPORT_FUNCTION(data_item, 1, true);
    EXPORT_FUNCTION(data_new, 0, false);
    

    EXPORT_FUNCTION(setPrimitivePrototype, 2, false);
    EXPORT_FUNCTION(object_slots, 1, false);
    EXPORT_FUNCTION(object_slotValues, 1, false);
    EXPORT_FUNCTION(object_equals, 1, true);
    
    EXPORT_FUNCTION(stringFromFile, 1, false);
	EXPORT_FUNCTION(stringFromInput, 1, false);
//    EXPORT_FUNCTION(stringFromLibrary, 0);
    EXPORT_FUNCTION(fileExists, 1, false);
    
	EXPORT_FUNCTION(charFromNumber, 1, false);
	EXPORT_FUNCTION(numberFromChar, 1, false);
    
    EXPORT_FUNCTION(boolYes, 0, false);
	EXPORT_FUNCTION(boolNo, 0, false);
    EXPORT_FUNCTION(error, 1, false);
    
    EXPORT_FUNCTION(getParameterObject, 0, false)
    
	return false;
}